from __future__ import unicode_literals
__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "oneID-connect"
__summary__ = ("Neustar TDI is an identity & authentication "
               "framework for the Internet of Things (IoT) "
               "devices, servers, and end-users.")

__uri__ = "https://github.com/OneID/oneID-connect-python"

__version__ = "0.20.0"

__author__ = "Neustar Inc."
__email__ = "support@oneID.com"

__license__ = "Apache License, Version 2.0"
__copyright__ = "Copyright 2012-2016 {0}".format(__author__)
